
/**************************************************************************
 *                                                                        *
 * This code is developed by Adam Li.  This software is an                *
 * implementation of a part of one or more MPEG-4 Video tools as          *
 * specified in ISO/IEC 14496-2 standard.  Those intending to use this    *
 * software module in hardware or software products are advised that its  *
 * use may infringe existing patents or copyrights, and any such use      *
 * would be at such party's own risk.  The original developer of this     *
 * software module and his/her company, and subsequent editors and their  *
 * companies (including Project Mayo), will have no liability for use of  *
 * this software or modifications or derivatives thereof.                 *
 *                                                                        *
 * Project Mayo gives users of the Codec a license to this software       *
 * module or modifications thereof for use in hardware or software        *
 * products claiming conformance to the MPEG-4 Video Standard as          *
 * described in the Open DivX license.                                    *
 *                                                                        *
 * The complete Open DivX license can be found at                         *
 * http://www.projectmayo.com/opendivx/license.php .                      *
 *                                                                        *
 **************************************************************************/

/**************************************************************************
 *
 *  encore.c, MPEG-4 Encoder Core Engine
 *
 *  Copyright (C) 2001  Project Mayo
 *
 *  Adam Li
 *
 *  DivX Advance Research Center <darc@projectmayo.com>
 *
 **************************************************************************/

/* This file contains encore(), which is the main entrance function for   */
/* Encoder Core Engine. It also contains the functions to initialize the  */
/* parameters, and redirect the output stream                             */

#include "encore.h"

#include "vop_code.h"
#include "rgb2yuv.h"
#include "text_dct.h"
#include "rc_q2.h"
#include "mom_bitstream.h"

typedef struct _REFERENCE
{
	unsigned long handle;
	float framerate;
	long bitrate;
	long rc_period;
	int x_dim, y_dim;
	int prev_rounding;
	int search_range;
	int max_quantizer;
	int min_quantizer;
	int flip;

	long seq;

	Vop *current;        /* the current frame to be encoded */
	Vop *reference;      /* the reference frame - reconstructed previous frame */
	Vop *reconstruct;    /* intermediate reconstructed frame - used in inter */
	Vop *error;          /* intermediate error frame - used in inter to hold prediction error */

	struct _REFERENCE *pnext;
} REFERENCE;

unsigned char *outStream;
long *outCount;
FILE *ftrace = NULL;
int max_quantizer, min_quantizer;

/* private functions used only in this file */
void init_vol_config(VolConfig *vol_config);
void init_vop(Vop *vop);
void InitBitCounts(BitCount *pnum_bits);
Int get_fcode (Int sr);


int encore(unsigned long handle, unsigned long enc_opt, void *param1, void *param2)
{
	static REFERENCE *ref = NULL;
	static VolConfig *vol_config;
	// a link list to keep the reference frame for all instances

	REFERENCE *ref_curr, *ref_last = NULL;
	int x_dim, y_dim, size;
	Vop *curr;
	BitCount num_bits;

	ref_curr = ref_last = ref;
	while (ref_curr != NULL)
	{
		if (ref_curr->handle == handle) break;
		ref_last = ref_curr;
		ref_curr = ref_last->pnext;
	}
	// create a reference for the new handle when no match is found
	if (ref_curr == NULL)
	{
		if (enc_opt & ENC_OPT_RELEASE) return ENC_OK;
		ref_curr = (REFERENCE *)malloc(sizeof(REFERENCE));
		ref_curr->handle = handle;
		ref_curr->seq = 0;
		ref_curr->pnext = NULL;
		if (ref) ref_last->pnext = ref_curr;
		else ref = ref_curr;
	}
	// initialize for a handle if requested
	if (enc_opt & ENC_OPT_INIT)
	{
		init_fdct_enc();
		init_idct_enc();

		// initializing rate control
		ref_curr->framerate = ((ENC_PARAM *)param1)->framerate;
		ref_curr->bitrate = ((ENC_PARAM *)param1)->bitrate;
		ref_curr->rc_period = ((ENC_PARAM *)param1)->rc_period;
		ref_curr->x_dim = ((ENC_PARAM *)param1)->x_dim;
		ref_curr->y_dim = ((ENC_PARAM *)param1)->y_dim;
		ref_curr->search_range = ((ENC_PARAM *)param1)->search_range;
		ref_curr->max_quantizer = ((ENC_PARAM *)param1)->max_quantizer;
		ref_curr->min_quantizer = ((ENC_PARAM *)param1)->min_quantizer;
		ref_curr->flip = ((ENC_PARAM *)param1)->flip;

		ref_curr->current = AllocVop(ref_curr->x_dim, ref_curr->y_dim);
		ref_curr->reference = AllocVop(ref_curr->x_dim + 2 * 16, 
			ref_curr->y_dim + 2 * 16);
		ref_curr->reconstruct = AllocVop(ref_curr->x_dim, ref_curr->y_dim);
		ref_curr->error = AllocVop(ref_curr->x_dim, ref_curr->y_dim);
		init_vop(ref_curr->current);
		init_vop(ref_curr->reference);
		init_vop(ref_curr->reconstruct);
		init_vop(ref_curr->error);
		ref_curr->reference->hor_spat_ref = -16;
		ref_curr->reference->ver_spat_ref = -16;
		SetConstantImage(ref_curr->reference->y_chan, 0);

		vol_config = (VolConfig *)malloc(sizeof(VolConfig));
		init_vol_config(vol_config);
		vol_config->frame_rate = ref_curr->framerate;
		vol_config->bit_rate = ref_curr->bitrate;
		RCQ2_init(22, vol_config, ref_curr->rc_period);

#ifdef _RC_
		ftrace = fopen("trace.txt", "w");
#endif
		return ENC_OK;
	}
	// release the reference associated with the handle if requested
	if (enc_opt & ENC_OPT_RELEASE)
	{
		RCQ2_Free();
		if (ref_curr == ref) ref = NULL;
		else ref_last->pnext = ref_curr->pnext;

		if (ref_curr->current) FreeVop(ref_curr->current);
		if (ref_curr->reference) FreeVop(ref_curr->reference);
		if (ref_curr->reconstruct) FreeVop(ref_curr->reconstruct);
		if (ref_curr->error) FreeVop(ref_curr->error);

		free(ref_curr);
		free(vol_config);
		if (ftrace) {
			fclose(ftrace);
			ftrace = NULL;
		};
		return ENC_OK;
	}

#ifdef _RC_
	fflush(ftrace);
	fprintf(ftrace, "\nNow coding frame # %d.\n", ref_curr->seq);
#endif

	// initialize the parameters (need to be cleaned later)
	InitBits(22, 5);

	max_quantizer = ref_curr->max_quantizer;
	min_quantizer = ref_curr->min_quantizer;

	x_dim = ref_curr->x_dim;
	y_dim = ref_curr->y_dim;
	size = x_dim * y_dim;
	outStream = (unsigned char *)(((ENC_FRAME *)param1)->bitstream);
	outCount = (long *)&(((ENC_FRAME *)param1)->length);
	*outCount = 0;

	curr = ref_curr->current;
	curr->width = x_dim;
	curr->height = y_dim;
	curr->sr_for = ref_curr->search_range;
	curr->fcode_for = get_fcode(curr->sr_for);

	// do RGB to YUV transformation for the input image
	RGB2YUV(x_dim, y_dim, ((ENC_FRAME *)param1)->bmp,
		curr->y_chan->f, curr->u_chan->f, curr->v_chan->f, ref_curr->flip);

	// adjust the rounding_type for the current image
	curr->rounding_type = 1 - ref_curr->prev_rounding;
	InitBitCounts(&num_bits);

	if (ref_curr->seq == 0) curr->prediction_type = I_VOP;
	else curr->prediction_type = P_VOP;

#ifdef _RC_
	fprintf(ftrace, "\nCoding frame #%d\n", ref_curr->seq);
#endif

	// Code the image data (YUV) of the current image
	VopCode(curr,
		ref_curr->reference,
		ref_curr->reconstruct,
		ref_curr->error,
		22, //vo_id,
		5, //vol_id,
		1, //enable_8x8_mv,
		(float)ref_curr->seq/ref_curr->framerate,  // time
		vol_config, 
		&num_bits);

	// update the rate control parameters
	RCQ2_Update2OrderModel(22, 5, // vo_id, vol_id,
		num_bits.vop,
		0, // num_bits_header set to 0
		0, // num_bits.psnr_y is not used, so set to 0
		GetVopPredictionType(curr));

	ref_curr->prev_rounding = curr->rounding_type;
	ref_curr->seq ++;

	if (curr->prediction_type == I_VOP)
		((ENC_RESULT *)param2)->isKeyFrame = 1;
	else 
		((ENC_RESULT *)param2)->isKeyFrame = 0;

	CloseBits(22, 5);

	return ENC_OK;
}


void init_vol_config(VolConfig *vol_config)
{
	/* configure VOL */
	vol_config->ident = 5;
	vol_config->M = 1;
	vol_config->frame_skip = 1;
	vol_config->quantizer = 8;
	vol_config->intra_quantizer = 8;
	vol_config->modulo_time_base[0] =0;
	vol_config->modulo_time_base[1] =0;
	vol_config->frame_rate = 30;
	vol_config->bit_rate = 800000;
}


void init_vop(Vop *vop)
{
	/* initialize VOPs */
	vop->vo_id = 22;
	vop->ident = 5;

	vop->quant_precision = 5;
	vop->bits_per_pixel = 8;
	vop->time_increment_resolution = 15 ;
	vop->intra_acdc_pred_disable = 0;
	vop->intra_dc_vlc_thr = 0;

	vop->sr_for = 512;
	vop->fcode_for = get_fcode(512);

	vop->y_chan->type = SHORT_TYPE;
	vop->u_chan->type = SHORT_TYPE;
	vop->v_chan->type = SHORT_TYPE;

	vop->time_increment_resolution = 15 ;

	vop->hor_spat_ref = 0;
	vop->ver_spat_ref = 0;

}

Void
InitBitCounts(BitCount *pnum_bits)
{
	pnum_bits->vol = 0;
	pnum_bits->vop = 0;
	pnum_bits->syntax = 0;
	pnum_bits->texture = 0;
	pnum_bits->motion = 0;
	pnum_bits->mot_shape_text = 0;

	return;
}

void PutByteStream(unsigned char byte)
{
	*(outStream ++) = byte;
	(*outCount) ++;
}

Int get_fcode (Int sr)
{
	if (sr<=16) return 1;
	else if (sr<=32) return 2;
	else if (sr<=64) return 3;
	else if (sr<=128) return 4;
	else if (sr<=256) return 5;
	else if (sr<=512) return 6;
	else if (sr<=1024) return 7;
	else return (-1);
}


